import os.path

from sunpy.data.test import rootdir as testrootdir

waveunitdir = os.path.join(testrootdir, 'waveunit')
MEDN_IMAGE = os.path.join(waveunitdir, 'medn_halph_fl_20050501_074655.fts')
MQ_IMAGE = os.path.join(waveunitdir, 'mq130812.084253.fits')
NA_IMAGE = os.path.join(waveunitdir, 'na120701.091058.fits')
SVSM_IMAGE = os.path.join(waveunitdir, 'svsm_e3100_S2_20110625_1856.fts')
